"""
Create Python packages just like I like them, without the hassle.
"""
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
import os
from pathlib import Path
import re
from subprocess import check_call
from typing import NamedTuple, Generator, List, Optional, Sequence, Tuple

from boxfactory.module import ImportType, Import, Module


TEST_DEPENDENCIES = ("coverage", "nose")
FORMATTER = "black"
LINTER = "flake8"
TYPER = "mypy"

SourceControlManager = Enum("SourceControlManager", ("GIT",))


class Version(NamedTuple):
    """
    A representation of a python versing number.
    """

    major: str
    minor: str

    def numeric(self) -> Tuple[int, int]:
        """
        The version as an interger
        """
        return (int(self.major), int(self.minor))

    @classmethod
    def from_string(cls, raw: str) -> "Version":
        """
        Build a Version from a string.
        """
        match = re.search(r"^(\d+)\.(\d+)$", raw)

        if match is None:
            raise ValueError(f"expected <major>.<minor>, received {raw}")

        return Version(*match.groups())


class ConsoleScript(NamedTuple):
    """
    Information about an entrypoint for a console script.
    """

    script: str
    modules: Sequence[str]
    function: str

    def __str__(self) -> str:
        modules = ".".join(self.modules)
        return f"{self.script} = {modules}:{self.function}"


def build(
    name: str,
    directory: Path,
    versions: Sequence[Version],
    description: Optional[str] = None,
    dependencies: Optional[Sequence[str]] = None,
    test_dependencies: Optional[Sequence[str]] = None,
    scripts: Optional[Sequence[ConsoleScript]] = None,
    scm: Optional[SourceControlManager] = None,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
    github_user: Optional[str] = None,
    license: Optional[str] = None,
    skip_tests: bool = False,
    skip_coverage: bool = False,
    skip_formatting: bool = False,
    skip_types: bool = False,
) -> List[Path]:
    """
    Build a package. Returns the list of created files.

    name: The name of the package.
    directory: The directory to create the package in.
    versions: Any versions of python this package supports (and should
        be tested against).
    description: A description of the package.
    dependencies: A list of packages the package depends upon.
    test_dependencies: A list of packages the package's tests depend
        upon. All package dependencies and nose will automatically be
        included.
    scripts: Any number of console scripts to create.
    scm: The Source control manager to use.
    author: The author of the package.
    author_email: The email address of the author.
    github_user: The name of the github account hosting this package.
    license: The license this package is provided under.
    skip_tests: If true, no tox file or test directory will be created.

    The following are ignored if skip_tests is True:

    skip_coverage: If true, tests will not include coverage checks.
    skip_formatting: If true, a formatter tox environment won't be
        created.
    skip_types: If true, a mypy tox environment won't be created.
    """
    if scripts is None:
        scripts = ()

    if test_dependencies is None:
        if skip_coverage:
            test_dependencies = ("nose",)
        else:
            test_dependencies = ("coverage", "nose")
    elif not skip_tests:
        test_dependencies = list(test_dependencies)

        if not skip_coverage and "coverage" not in test_dependencies:
            test_dependencies.append("coverage")

        if "nose" not in test_dependencies:
            test_dependencies.append("nose")

    base_directory = directory / name
    base_directory.mkdir()

    created = build_package(
        name, base_directory, description=description, scripts=scripts
    )

    created.extend(
        build_requirements(
            base_directory,
            dependencies=dependencies,
            test_dependencies=test_dependencies,
            skip_tests=skip_tests,
            skip_coverage=skip_coverage,
            skip_formatting=skip_formatting,
            skip_types=skip_types,
        )
    )

    if not skip_tests:
        created.extend(
            build_tests(
                name,
                base_directory,
                versions,
                skip_coverage=skip_coverage,
                skip_formatting=skip_formatting,
                skip_types=skip_types,
            )
        )

    created.extend(
        build_setup(
            name,
            base_directory,
            versions,
            description=description,
            dependencies=dependencies,
            test_dependencies=test_dependencies,
            scripts=scripts,
            author=author,
            author_email=author_email,
            github_user=github_user,
            license=license,
            skip_tests=skip_tests,
            skip_coverage=skip_coverage,
        )
    )

    if scm == SourceControlManager.GIT:
        initialize_git_repository(base_directory, created)

    return created


def build_package(
    name: str,
    directory: Path,
    description: Optional[str] = None,
    scripts: Optional[Sequence[ConsoleScript]] = None,
):
    """
    Build the actual package. Returns the list of created files.

    name: The name of the package.
    directory: The directory to create the package in.
    description: A description of the package.
    scripts: Any number of console scripts to create.
    """
    root = Module(name, description, force_init=True)

    for script in scripts or ():
        if root.module != script.modules[0]:
            raise ValueError(script.modules)

        script_name = script.script
        function_name = script.function

        docstring = f"Run {script_name} from the command line."
        argparse = Import(
            ImportType.STANDARD_LIBRARY, ("argparse",), "ArgumentParser"
        )
        function_body = (
            f"def {function_name}(input_args=None):\n"
            f'    """\n'
            f"    {docstring}\n"
            f'    """\n'
            f'    parser = ArgumentParser(description="Run {script_name}")\n'
            f"    args = parser.parse_args(args=input_args)\n"
            f"    raise NotImplementedError(args)"
        )

        node = root
        for module in script.modules[1:]:
            child = node.submodules.get(module)

            if child is None:
                node.submodules[module] = child = Module(module)

            node = child

        node.update(
            docstring=docstring,
            imports=(argparse,),
            functions={function_name: function_body},
        )

    return root.generate(directory)


def build_requirements(
    directory: Path,
    dependencies: Optional[Sequence[str]] = None,
    test_dependencies: Optional[Sequence[str]] = None,
    skip_tests: bool = False,
    skip_coverage: bool = False,
    skip_formatting: bool = False,
    skip_types: bool = False,
) -> List[Path]:
    """
    Build the requirements files.

    directory: The base package directory.
    dependencies: (optional, None) a list of package dependencies.
    test_dependencies: (optional, None) a list of test dependencies.
    skip_tests: If true, no tox file or test directory will be created.

    The following are ignored if skip_tests is True:

    skip_coverage: If true, tests will not include coverage checks.
    skip_formatting: If true, a formatter tox environment won't be
        created.
    skip_types: If true, a mypy tox environment won't be created.
    """
    created = []

    requirements = directory / "requirements"

    requirements.mkdir()

    install = requirements / "install.txt"
    created.append(install)

    if dependencies:
        install.write_text("\n".join(dependencies))
    else:
        install.touch()

    if not skip_tests:
        test = requirements / "tests.txt"
        created.append(test)

        if test_dependencies:
            test_dependencies = list(test_dependencies)

            for package in TEST_DEPENDENCIES:
                if package not in test_dependencies:
                    test_dependencies.append(package)
        else:
            test_dependencies = TEST_DEPENDENCIES

        test.write_text("\n".join(test_dependencies))

        if not skip_formatting:
            formatting = requirements / "formatting.txt"
            created.append(formatting)

            formatting.write_text(f"{FORMATTER}\n{LINTER}")

        if not skip_types:
            types = requirements / "types.txt"
            created.append(types)

            types.write_text(TYPER)

    return created


def build_tests(
    name: str,
    directory: Path,
    versions: Sequence[Version],
    skip_coverage: bool = False,
    skip_formatting: bool = False,
    skip_types: bool = False,
) -> List[Path]:
    """
    Build the requirements files.

    name: The name of the package.
    directory: The base package directory.
    versions: Any versions of python this package supports (and should
        be tested against).
    skip_coverage: If true, tests will not include coverage checks.
    skip_formatting: If true, a formatter tox environment won't be
        created.
    skip_types: If true, a mypy tox environment won't be created.
    """
    tox = directory / "tox.ini"

    created = [tox]

    version_envs = tuple(
        f"py{version.major}{version.minor}" for version in versions
    )

    with tox.open("w") as stream:
        stream.write("[tox]\nenvlist = ")
        stream.write(", ".join(version_envs))

        if not skip_formatting:
            stream.write(", formatting")

        if not skip_types:
            stream.write(", types")

        stream.write("\n")

        for env in version_envs:
            stream.write(
                f"\n[testenv:{env}]\n"
                "deps =\n"
                "    -rrequirements/install.txt\n"
                "    -rrequirements/tests.txt\n"
                "commands =\n"
            )

            if skip_coverage:
                stream.write("    nosetests tests -s\n")
            else:
                # quick aside: show missing lines
                rc = directory / ".coveragerc"
                created.append(rc)
                rc.write_text("[report]\nshow_missing = True\n")

                stream.write(
                    "    nosetests tests -s --with-coverage --cover-erase "
                    f"--cover-inclusive --cover-package={name}\n"
                )

        if not skip_formatting:
            formatter_parts = [
                f"    {FORMATTER} {name} tests --check --diff --line-length 79"
            ]

            if all(version.numeric() >= (3, 6) for version in versions):
                formatter_parts.append(" --py36")

            stream.write(
                "\n[testenv:formatting]\n"
                "deps =\n"
                "    -rrequirements/install.txt\n"
                "    -rrequirements/formatting.txt\n"
                "commands =\n"
                f"{''.join(formatter_parts)}\n"
                f"    {LINTER} {name} tests\n"
            )

        if not skip_types:
            stream.write(
                "\n[testenv:types]\n"
                "deps =\n"
                "    -rrequirements/install.txt\n"
                "    -rrequirements/types.txt\n"
                f"commands = {TYPER} {name}\n"
            )

    tests = directory / "tests"
    tests.mkdir()
    testfile = tests / f"test_{name}.py"
    testfile.touch()
    created.append(testfile)

    return created


def build_setup(
    name: str,
    directory: Path,
    versions: Sequence[Version],
    description: Optional[str] = None,
    dependencies: Optional[Sequence[str]] = None,
    test_dependencies: Optional[Sequence[str]] = None,
    scripts: Optional[Sequence[ConsoleScript]] = None,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
    github_user: Optional[str] = None,
    license: Optional[str] = None,
    skip_tests: bool = False,
    skip_coverage: bool = False,
):
    """
    Build the setup files. Returns the list of created files.

    name: The name of the package.
    directory: The directory to create the package in.
    versions: Any versions of python this package supports (and should
        be tested against).
    description: A description of the package.
    dependencies: A list of packages the package depends upon.
    test_dependencies: A list of packages the package's tests depend
        upon. All package dependencies and nose will automatically be
        included.
    scripts: Any number of console scripts to create.
    author: The author of the package.
    author_email: The email address of the author.
    github_user: The name of the github account hosting this package.
    license: The license this package is provided under.
    skip_tests: If true, no tox file or test directory will be created.
    skip_coverage: If true, tests will not include coverage checks.
    """
    created = []

    setup = directory / "setup.py"
    created.append(setup)

    with setup.open("w") as stream:
        stream.write(
            '"""\n'
            "To install:\n"
            "\n"
            "    python setup.py install\n"
            '"""\n'
            "from pathlib import Path\n"
            "from setuptools import setup\n"
            "\n\n"
            "setup(\n"
            f"    name={name!r},\n"
            f"    description=(\n        {description!r}\n    ),\n"
            "    long_description=("
            "Path(__file__).parent / 'README.md').read_text('utf-8'),\n"
            "    version='0.1.0',\n"
        )

        if author is not None:
            stream.write(f"    author={author!r},\n")

        if author_email is not None:
            stream.write(f"    author_email={author_email!r},\n")

        if github_user is not None:
            stream.write(
                f"    url='https://github.com/{github_user}/{name}',\n"
            )

        stream.write(
            f"    license={license!r},\n"
            "    packages=(\n"
            f"        {name!r},\n"
            "    ),\n"
        )

        if scripts:
            stream.write(
                "    entry_points={\n" "        'console_scripts': (\n"
            )

            for script in scripts:
                stream.write(f"            '{script}',\n")

            stream.write("        )\n    },\n")

        if dependencies:
            stream.write("    install_requires=(\n")

            for dependency in dependencies:
                stream.write(f"        {dependency!r},\n")

            stream.write("    ),\n")

        if test_dependencies:
            stream.write("    tests_require=(\n")

            for dependency in test_dependencies:
                stream.write(f"        {dependency!r},\n")

            stream.write("    ),\n")

        stream.write(
            "    classifiers=(\n"
            '        "Development Status :: 2 - Pre-Alpha",\n'
        )

        major_versions = set()

        classifier_versions = []

        for version in versions:
            major_versions.add(str(version.major))

            classifier_versions.append(
                "Programming Language :: Python :: "
                f"{version.major}.{version.minor}"
            )

        if len(major_versions) == 1:

            classifier_versions.append(
                "Programming Language :: Python :: "
                f"{next(iter(major_versions))} :: Only"
            )

        classifier_versions.extend(
            (
                "Programming Language :: Python :: " f"{major_version}"
                for major_version in major_versions
            )
        )

        version_strings = "".join(
            f'        "{version}",\n'
            for version in sorted(classifier_versions)
        )
        stream.write(f"{version_strings}    ),\n)\n")

    if not skip_tests:
        setup_config = directory / "setup.cfg"
        created.append(setup_config)

        with setup_config.open("w") as stream:
            stream.write(
                "[nosetests]\n" "verbosity = 2\n" "detailed-errors = 1\n"
            )

            if skip_coverage:
                stream.write("with-coverage = 0\n")
            else:
                stream.write(f"with-coverage = 1\ncover-package = {name}\n")

    manifest = directory / "MANIFEST.in"
    created.append(manifest)
    manifest.write_text("include *.txt\ninclude *.md\n")

    changes = directory / "CHANGES.txt"
    created.append(changes)
    now = datetime.now()
    changes.write_text(now.strftime("v0.0.0, %Y/%m/%d -- Package created."))

    readme = directory / "README.md"
    created.append(readme)

    with readme.open("w") as stream:
        stream.write(f"# {name}\n\n")

        if description:
            stream.write(f"{description}\n\n")

        stream.write(
            "## Installation\n\n```sh\npython setup.py install\n```\n"
        )

    return created


@contextmanager
def working_directory(directory: Path) -> Generator[None, None, None]:
    """
    Change the working directory for the duration of a context manager.
    """
    cwd = Path.cwd()

    try:
        os.chdir(directory)

        yield
    finally:
        os.chdir(cwd)


def initialize_git_repository(directory: Path, files: List[Path]):
    """
    Make the package a git repository.

    directory: The package directory.
    files: The files that should be added to the repository.
    """
    gitignore = directory / ".gitignore"
    files.append(gitignore)

    gitignore.write_text(
        ".coverage\n"
        ".DS_Store\n"
        ".mypy_cache\n"
        ".tox\n"
        "*.egg-info\n"
        "*.py[cod]\n"
        "__pycache__/\n"
        "build/\n"
        "dist/\n"
    )

    absolutes = [file.absolute() for file in files]

    check_call(("git", "init", str(directory)))

    cwd = Path.cwd()

    try:
        os.chdir(directory)

        for file in absolutes:
            check_call(("git", "add", str(file)))

        check_call(("git", "commit", "--message", "Package created."))
    finally:
        os.chdir(cwd)
