"""
The boxfactory utility
"""
from argparse import ArgumentParser
from pathlib import Path
from sys import version_info

from boxfactory import ConsoleScript, SourceControlManager, Version, build


def main(input_args=None):
    """
    Run the boxfactory script.
    """
    parser = ArgumentParser(description="Build Python packages")
    parser.add_argument("name", help="The name of the package")
    parser.add_argument(
        "location",
        nargs="?",
        type=Path,
        default=Path.cwd(),
        help="Where to create the package",
    )
    parser.add_argument(
        "--description", "-d", help="A short description of the package."
    )
    parser.add_argument(
        "--dependencies", nargs="+", help="Any number of package dependencies."
    )
    parser.add_argument(
        "--test-dependencies",
        nargs="+",
        help="Any number of package dependencies needed for testing.",
    )
    parser.add_argument(
        "--versions",
        "-v",
        nargs="+",
        type=Version.from_string,
        help="A version of python the package supports (e.g., 3.6).",
    )
    script_group = parser.add_mutually_exclusive_group()
    script_group.add_argument(
        "--cli",
        "-c",
        action="store_true",
        help=(
            "Add a console script entrypoint named <package name> that "
            "points to <package>.cli:main."
        ),
    )
    script_group.add_argument(
        "--script",
        "-s",
        nargs="+",
        type=str,
        action="append",
        help=(
            "An (script, *mudules, method) tuple. "
            "Can be supplied multiple times."
        ),
    )
    parser.add_argument(
        "--author", "-a", type=str, help="The name of the package author."
    )
    parser.add_argument(
        "--email", "-e", type=str, help="The email of the package author."
    )
    parser.add_argument(
        "--github-user",
        "-u",
        type=str,
        help="The github username of the author (used for setup.py url).",
    )
    parser.add_argument(
        "--license",
        type=str,
        help="The license the package is provided under.",
    )
    parser.add_argument(
        "--git",
        "-g",
        action="store_true",
        help="Initialize the package directory as a git repository.",
    )
    parser.add_argument(
        "--skip-tests", action="store_true", help="Don't add tests."
    )
    parser.add_argument(
        "--skip-coverage", action="store_true", help="Don't add test coverage."
    )
    parser.add_argument(
        "--skip-formatting",
        action="store_true",
        help="Don't add formatting tests.",
    )
    parser.add_argument(
        "--skip-types", action="store_true", help="Don't add typing tests."
    )

    args = parser.parse_args(args=input_args)

    if args.versions is None:
        args.versions = (
            Version(str(version_info.major), str(version_info.minor)),
        )

    if args.cli:
        scripts = (ConsoleScript(args.name, (args.name, "cli"), "main"),)
    else:
        scripts = []

        if args.script:
            for script, *modules, function in args.script:
                scripts.append(ConsoleScript(script, modules, function))

    if args.git:
        scm = SourceControlManager.GIT
    else:
        scm = None

    files = build(
        name=args.name,
        directory=args.location,
        versions=args.versions,
        description=args.description,
        dependencies=args.dependencies,
        test_dependencies=args.test_dependencies,
        scripts=scripts,
        author=args.author,
        author_email=args.email,
        github_user=args.github_user,
        license=args.license,
        scm=scm,
        skip_tests=args.skip_tests,
        skip_coverage=args.skip_coverage,
        skip_formatting=args.skip_formatting,
        skip_types=args.skip_types,
    )

    for file in sorted(files):
        print(file)


if __name__ == "__main__":
    main()
