"""
A representation of a python module.
"""
# This whole module is poorly over-engineered and I should have just
# not done any of this.
from enum import IntEnum
from functools import reduce
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence, Tuple


MAX_LINE_LENGTH = 79


ImportType = IntEnum(
    "ImportType", ("FUTURE", "STANDARD_LIBRARY", "THIRD_PARTY", "PACKAGE")
)


class Import:
    """
    An import from a module.
    """

    def __init__(
        self,
        import_type: ImportType,
        module: Sequence[str],
        *items: str,
        import_base: bool = False,
        import_as: Optional[str] = None,
    ) -> None:
        self.type = import_type
        self.module = tuple(module)
        self.items = items
        self.import_base = import_base
        self.import_as = import_as

        if not (items or import_base):
            raise ValueError(self)

    def merge(self, other: "Import") -> "Import":
        """
        Combine two imports (of the same module) together.
        """
        if self.type != other.type or self.module != other.module:
            raise ValueError((self, other))

        if self.items:
            if other.items:
                # we want to preserve order so we can't just union them.
                new_items = set(other.items) - set(self.items)

                items = (
                    *self.items,
                    *(item for item in other.items if item in new_items),
                )
            else:
                items = self.items
        else:
            items = other.items

        if self.import_base:
            if other.import_base and self.import_as != other.import_as:
                raise NotImplementedError(
                    "There is probably a good reason why you need to "
                    "import this module under two names, but you're on "
                    "you're own."
                )

            import_base = True
            import_as = self.import_as
        else:
            import_base = other.import_base
            import_as = other.import_as

        return Import(
            self.type,
            self.module,
            *items,
            import_base=import_base,
            import_as=import_as,
        )

    def format(
        self,
        sort_function: Optional[
            Callable[[Sequence[str]], Sequence[str]]
        ] = None,
    ) -> str:
        """
        Format the import as a PEP 8-formated import statement.

        sort_function: (optional, None) A function for ordering the
        items being imported.
        """
        module, *submodules = self.module
        module_length = len(module)
        num_submodules = len(submodules)

        parts: List[str] = []

        if self.import_base:
            parts.extend(("import ", module))
            line_length = 7 + module_length

            for index, submodule in enumerate(submodules):
                if self.import_as or index + 1 < num_submodules:
                    buffer = 3
                else:
                    buffer = 1

                name_length = len(submodule)

                if line_length + name_length + buffer > MAX_LINE_LENGTH:
                    parts.extend((" \\\n    .", submodule))
                    line_length = 5 + name_length
                else:
                    parts.extend((".", submodule))
                    line_length += 1 + name_length

            if self.import_as:
                if line_length + 5 > MAX_LINE_LENGTH:
                    parts.extend((" \\\n    as ", self.import_as))
                else:
                    parts.append(" as ")
                    line_length += 4

                    if line_length + len(self.import_as) > MAX_LINE_LENGTH:
                        parts.append("\\\n    ")

                    parts.append(self.import_as)

            parts.append("\n")
            line_length = 0

        if self.items:
            parts.extend(("from ", module))
            line_length = 5 + module_length

            for submodule in submodules:
                name_length = len(submodule)

                if line_length + name_length + 3 > MAX_LINE_LENGTH:
                    parts.extend((" \\\n    .", submodule))
                    line_length = 5 + name_length
                else:
                    parts.extend((".", submodule))
                    line_length += 1 + name_length

            if line_length + 9 > MAX_LINE_LENGTH:
                parts.append(" \\\n    import ")
                line_length = 11
            else:
                parts.append(" import ")
                line_length += 8

            items: Sequence[str] = self.items

            if sort_function:
                items = sort_function(items)

            num_items = len(items)
            min_length = (
                1
                + (2 * (num_items - 1))
                + reduce(lambda length, item: length + len(item), items, 0)
            )

            if line_length + min_length < MAX_LINE_LENGTH:
                parts.extend((", ".join(items), "\n"))
            else:
                first, *remaining = items

                parts.extend(("(\n    ", first))
                line_length = 4 + len(first)

                for index, item in enumerate(remaining, 1):
                    if index + 1 < num_items:
                        buffer = 3
                    else:
                        buffer = 2

                    item_length = len(item)

                    if line_length + item_length + buffer > MAX_LINE_LENGTH:
                        parts.extend((",\n    ", item))
                        line_length = 4 + item_length
                    else:
                        parts.extend((", ", item))
                        line_length += 2 + item_length

                parts.append("\n)\n")

        return "".join(parts)


class Module:
    """
    A representation of a python module
    """

    def __init__(
        self,
        module: str,
        docstring: Optional[str] = None,
        imports: Optional[Sequence[Import]] = None,
        functions: Optional[Dict[str, str]] = None,
        force_init: bool = False,
    ) -> None:
        self.module = module
        self.docstring = docstring
        self.imports: Dict[Tuple[str, ...], Import] = {}
        self.functions = functions or {}
        self.submodules: Dict[str, Module] = {}
        self.force_init = force_init

        if imports:
            for import_ in imports:
                self.imports[import_.module] = import_

    def update(
        self,
        *,
        docstring: Optional[str] = None,
        imports: Optional[Sequence[Import]] = None,
        functions: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Update the module to add more information.

        Note: docstring will be ignored if the module already has a
        docstring.
        """
        if docstring is not None and self.docstring is None:
            self.docstring = docstring

        updated_imports: Dict[Tuple[str, ...], Import] = {}
        updated_functions: Dict[str, str] = {}

        if imports:
            for import_ in imports:
                existing = self.imports.get(import_.module)

                if existing:
                    import_ = existing.merge(import_)

                updated_imports[import_.module] = import_

        if functions:
            for name, body in functions.items():
                if name in self.functions:
                    raise ValueError(name)

                updated_functions[name] = body

        for module, import_ in updated_imports.items():
            self.imports[module] = import_

        for name, body in updated_functions.items():
            self.functions[name] = body

    def generate(
        self,
        directory: Path,
        item_sort_function: Optional[
            Callable[[Sequence[str]], Sequence[str]]
        ] = None,
    ) -> List[Path]:
        """
        Generate the module, returning the list of all files created.

        directory: The path to the directory to create the module in.
        item_sort_function: A function to sort individual items in a
            from import.
        """
        files = []

        as_file = directory / f"{self.module}.py"
        as_directory = directory / self.module

        if self.force_init or self.submodules:
            if as_file.exists():
                raise IOError(as_file)

            as_directory.mkdir()

            path = as_directory / "__init__.py"
        else:
            if as_directory.exists():
                raise IOError(as_directory)

            path = as_file

        files.append(path)

        with path.open("w") as stream:
            if self.docstring:
                # TODO: this docstring may be too long
                stream.write(f'"""\n{self.docstring}\n"""\n')

            last_type = None

            for import_ in sorted(
                self.imports.values(),
                key=lambda import_: (import_.type, import_.module),
            ):
                if last_type and import_.type != last_type:
                    stream.write("\n")

                stream.write(import_.format(sort_function=item_sort_function))
                last_type = import_.type

            for function in self.functions.values():
                stream.write("\n\n")
                stream.write(function)
                stream.write("\n")

        for submodule in self.submodules.values():
            files.extend(submodule.generate(as_directory, item_sort_function))

        return files


def sort_items(items: Sequence[str]) -> List[str]:
    """
    Sort items based on their import name.
    """
    # NOTE: we have guarantees that each item will be unique, but not
    # that each import name will be unique (we won't catch "A, a as A"
    # or "a as A, b as A"). sorted is stable so we will at least
    # preserve that order.
    return sorted(items, key=lambda item: item.split(" as ", 1)[-1])


def group_import_items(items: Sequence[str]) -> Tuple[str, ...]:
    """
    Sort import items based on whether they are constants, classes, or
    functions.

    NOTE: This function assumes that constants will be in
    SCREAMING_SNAKE_CASE, classes will be in PascalCase, and functions
    will be in snake_case or camelCase. This is untrue of the standard
    library, and just doesn't make sense for names written in scripts
    that don't have letter case (or where the base grapheme isn't an
    alphabetic letter).

    All names that don't fall into any category (e.g., 寫, ලියන්න),
    or fit into multiple categories (e.g., _, _1) will be placed after
    all others.

    NOTE: PascalCase and CamelCase are being differentiated by first
    symbol which can have it's case altered (e.g., _HiddenClass will be
    considered a class). This might have weird results if a name starts
    with a number (e.g., _1TimeKey is read as a class, _2nd is read as a
    function). Those are hopefully rare anyway.
    """

    def get_name(item: str) -> str:
        """
        Get the actual name of the item.
        """
        return item.split(" as ", 1)[-1]

    constants: List[str] = []
    # There is a now-fixed bug in pycodestyle where it gets confused by
    # type annotations if the variable's name starts with a keyword (in
    # this case, 'class'). It has been patched, but that fix seemingly
    # hasn't bubbled up to pyflakes yet.
    classes = []  # type: List[str]
    functions: List[str] = []
    unknown: List[str] = []

    for item in items:
        name = get_name(item)
        upper = name.upper()
        lower = name.lower()

        group = unknown

        if upper == name:
            if lower != name:
                group = constants
        elif lower == name:
            group = functions
        else:
            # contains both uppercase and lowercase symbols. If the
            # first symbol with case is upper it's a class, if it's
            # lower it's a function.
            for n, u, l in zip(name, upper, lower):
                if n == l:
                    if n != u:
                        group = functions
                        break
                elif n == u:
                    group = classes
                    break

        group.append(item)

    return (
        *sort_items(constants),
        *sort_items(classes),
        *sort_items(functions),
        *sort_items(unknown),
    )
